@@ -21,17 +21,15 @@ import logging
 import re
 from functools import partial
 from html import unescape
-from operator import attrgetter, truth
-from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Set, Union
+from operator import truth
+from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Set
 
-import beets
-import beets.ui
-import beetsplug.fetchart as fetchart
 import requests
 import six
-from beets import plugins
-from beets.autotag.hooks import AlbumInfo, Distance, TrackInfo
-from beets.library import Item
+from beets import __version__, config, library, plugins
+from beets.autotag.hooks import AlbumInfo, TrackInfo
+from beets.importer import ImportTask
+from beetsplug import fetchart  # type: ignore[attr-defined]
 
 from ._metaguru import DATA_SOURCE, DEFAULT_MEDIA, Metaguru, urlify
 
@@ -49,7 +47,7 @@ DEFAULT_CONFIG: JSONDict = {
 SEARCH_URL = "https://bandcamp.com/search?q={0}&page={1}"
 ALBUM_URL_IN_TRACK = re.compile(r'inAlbum":{[^}]*"@id":"([^"]*)"')
 SEARCH_ITEM_PAT = 'href="(https://[^/]*/{}/[^?]*)'
-USER_AGENT = "beets/{} +http://beets.radbox.org/".format(beets.__version__)
+USER_AGENT = "beets/{} +http://beets.radbox.org/".format(__version__)
 ALBUM_SEARCH = "album"
 ARTIST_SEARCH = "band"
 TRACK_SEARCH = "track"
@@ -86,7 +84,8 @@ class BandcampRequestsHandler:
 class BandcampAlbumArt(BandcampRequestsHandler, fetchart.RemoteArtSource):
     NAME = "Bandcamp"
 
-    def get(self, album: AlbumInfo, plugin, paths) -> Iterator[fetchart.Candidate]:
+    def get(self, album, plugin, paths):
+        # type: (AlbumInfo, plugins.BeetsPlugin, List) -> Iterator[fetchart.Candidate]  # noqa
         """Return the url for the cover from the bandcamp album page.
         This only returns cover art urls for bandcamp albums (by id).
         """
@@ -111,33 +110,67 @@ class BandcampAlbumArt(BandcampRequestsHandler, fetchart.RemoteArtSource):
             self._info("Art cover is already present")
 
 
-class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
-    _gurucache: Dict[str, Metaguru]
+def _from_bandcamp(clue: str) -> bool:
+    return ".bandcamp." in clue or (
+        clue.startswith("http") and ("/album/" in clue or "/track/" in clue)
+    )
 
-    media: str
+
+class BandcampAdditionalData:
+    write: bool
     excluded_extra_fields: Set[str]
+    guru: Callable
+    _info: Callable
+
+    def verify_and_add(
+        self, item: library.Item, get_data: Callable[[str], str], excluded: Set[str]
+    ) -> None:
+        name = "Additional data"
+        for bandcamp_field in set(ADDITIONAL_DATA_MAP).difference(excluded):
+            item_field = ADDITIONAL_DATA_MAP[bandcamp_field]
+
+            if not getattr(item, item_field, None) or (
+                item_field == "comments" and item.comments.startswith("Visit http")
+            ):
+                new_value = get_data(bandcamp_field)
+                if new_value:
+                    item.update({item_field: new_value})
+                    self._info("{}: obtained {} for {}", name, bandcamp_field, item)
+                else:
+                    self._info("{}: {} not found for {}", name, bandcamp_field, item)
+            else:
+                self._info("{}: {}: already present on {}", name, item_field, item)
+        return item
+
+    def _add_additional_data(self, item: library.Item) -> None:
+        """If not excluded, fetch and store:
+        * lyrics
+        * release description as comments
+        """
+        get_data_call = partial(getattr, self.guru(item.mb_albumid or item.mb_trackid))
+        return self.verify_and_add(item, get_data_call, self.excluded_extra_fields)
+
+
+class BandcampPlugin(
+    BandcampRequestsHandler, plugins.BeetsPlugin, BandcampAdditionalData
+):
+    _gurucache: Dict[str, Metaguru]
 
     def __init__(self) -> None:
         super().__init__()
         self.config.add(DEFAULT_CONFIG.copy())
-        # ~~~ DEPRECATED
-        if not self.config["lyrics"]:
-            ADDITIONAL_DATA_MAP.pop("lyrics", None)
-        # ~~~
-        self.media = self.config["preferred_media"].as_str()
+
+        self.write = config["import"]["write"].get()
         self.excluded_extra_fields = set(self.config["exclude_extra_fields"].get())
-        self.import_stages = [self.imported]
+        self.register_listener("import_task_apply", self.add_additional_data)
         self.register_listener("pluginload", self.loaded)
         self._gurucache = dict()
 
-    @staticmethod
-    def _from_bandcamp(clue: Union[Item, str]) -> bool:
-        """Accepts either an item or the mb_artistid."""
-        if isinstance(clue, Item):
-            clue = clue.mb_albumid or clue.mb_trackid
-        return "bandcamp" in clue or (
-            clue.startswith("http") and ("album" in clue or "track" in clue)
-        )
+    def add_additional_data(self, task: ImportTask) -> None:
+        """Import hook for fetching additional data from bandcamp."""
+        for item in task.items:
+            if _from_bandcamp(item.mb_albumid or item.mb_trackid):
+                self._add_additional_data(item)
 
     def guru(self, url: str, html: Optional[str] = None) -> Optional[Metaguru]:
         """Return cached guru. If there isn't one, fetch the url if html isn't
@@ -149,74 +182,42 @@ class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
         if not html:
             html = self._get(url)
         if html:
-            self._gurucache[url] = Metaguru(html, self.media)
+            self._gurucache[url] = Metaguru(
+                html,
+                self.config["preferred_media"].as_str(),
+                self.config["include_digital_only_tracks"].get(),
+            )
         return self._gurucache.get(url)
 
-    def add_additional_data(self, item: Item, write: bool = False) -> None:
-        """Fetch and store:
-        * lyrics, if enabled
-        * release description as comments
-        """
-        guru = self.guru(item.mb_albumid or item.mb_trackid)
-
-        backup = ""
-        if item.comments.startswith == "Visit http":
-            backup = item.comments
-            item.comments = ""
-
-        for bandcamp_field, item_field in ADDITIONAL_DATA_MAP.items():
-            if item_field in self.excluded_extra_fields:
-                continue
-
-            if getattr(item, item_field, None):
-                self._info("{} field: already present on {}", item_field, item)
-                continue
-
-            setattr(item, item_field, getattr(guru, bandcamp_field))
-            if getattr(item, item_field, None):
-                self._info("Obtained {} for {}", bandcamp_field, item)
-        if not item.comments:
-            item.comments = backup
-
-        if write:
-            item.try_write()
-        item.store()
-
-    def imported(self, _: Any, task: Any) -> None:
-        """Import hook for fetching additional data from bandcamp."""
-        for item in task.imported_items():
-            if self._from_bandcamp(item):
-                self.add_additional_data(item, write=True)
-
     def loaded(self) -> None:
         """Add our own artsource to the fetchart plugin."""
         # TODO: This is ugly, but i didn't find another way to extend fetchart
         # without declaring a new plugin.
         if self.config["art"]:
+            fetchart.ART_SOURCES[DATA_SOURCE] = BandcampAlbumArt
+            fetchart.SOURCE_NAMES[BandcampAlbumArt] = DATA_SOURCE
+            bandcamp_fetchart = BandcampAlbumArt(self._log, self.config)
+
             for plugin in plugins.find_plugins():
                 if isinstance(plugin, fetchart.FetchArtPlugin):
-                    plugin.sources = [
-                        BandcampAlbumArt(plugin._log, self.config)
-                    ] + plugin.sources
-                    fetchart.ART_SOURCES[DATA_SOURCE] = BandcampAlbumArt
-                    fetchart.SOURCE_NAMES[BandcampAlbumArt] = DATA_SOURCE
+                    plugin.sources = [bandcamp_fetchart, *plugin.sources]
                     break
 
-    def _cheat_mode(self, item: Item, name: str, _type: str) -> Optional[str]:
+    def _cheat_mode(self, item: library.Item, name: str, _type: str) -> str:
         reimport_url: str = getattr(item, f"mb_{_type}id", "")
         if "bandcamp" in reimport_url:
             return reimport_url
 
-        if "Visit" in item.comments:
+        if item.comments.startswith("Visit"):
             match = re.search(r"https:[/a-z.-]+com", item.comments)
             if match:
-                url: str = "{}/{}/{}".format(match.group(), _type, urlify(name))
+                url = "{}/{}/{}".format(match.group(), _type, urlify(name))
                 self._info("Trying our guess {} before searching", url)
                 return url
-        return None
+        return ""
 
     def candidates(self, items, artist, album, va_likely, extra_tags=None):
-        # type: (List[Item], str, str, bool, Optional[JSONDict]) -> Iterator[AlbumInfo]
+        # type: (List[library.Item], str, str, bool, Optional[JSONDict]) -> Iterator[AlbumInfo]  # noqa
         """Return a sequence of AlbumInfo objects that match the
         album whose items are provided.
         """
@@ -225,10 +226,11 @@ class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
             initial_guess = self.get_album_info(initial_url) if initial_url else None
             if initial_guess:
                 return iter([initial_guess])
+
         return filter(truth, map(self.get_album_info, self._search(album, ALBUM_SEARCH)))
 
     def item_candidates(self, item, artist, title):
-        # type: (Item, str, str) -> Iterator[TrackInfo]
+        # type: (library.Item, str, str) -> Iterator[TrackInfo]
         """Return a sequence of TrackInfo objects that match the provided item.
         If the track was downloaded directly from bandcamp, it should contain
         a comment saying 'Visit <label-url>' - we look at this first by converting
@@ -238,36 +240,48 @@ class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
         initial_guess = self.get_track_info(initial_url) if initial_url else None
         if initial_guess:
             return iter([initial_guess])
+
         query = title or item.album or artist
         return filter(truth, map(self.get_track_info, self._search(query, TRACK_SEARCH)))
 
     def album_for_id(self, album_id: str) -> Optional[AlbumInfo]:
         """Fetch an album by its bandcamp ID."""
-        return self.get_album_info(album_id)
+        if _from_bandcamp(album_id):
+            return self.get_album_info(album_id)
+
+        self._info("Not a bandcamp URL, skipping")
+        return None
 
     def track_for_id(self, track_id: str) -> Optional[TrackInfo]:
         """Fetch a track by its bandcamp ID."""
-        return self.get_track_info(track_id)
+        if _from_bandcamp(track_id):
+            return self.get_track_info(track_id)
+
+        self._info("Not a bandcamp URL, skipping")
+        return None
+
+    def handle(self, guru: Metaguru, attr: str, _id: str) -> Any:
+        try:
+            return getattr(guru, attr)
+        except (KeyError, ValueError, AttributeError):
+            self._info("Failed obtaining {}", _id)
+            return None
+        except Exception:  # pylint: disable=broad-except
+            url = "https://github.com/snejus/beetcamp/issues/new"
+            self._exc("Unexpected error obtaining {}, please report at {}", _id, url)
+            return None
 
     def get_album_info(self, url: str) -> Optional[AlbumInfo]:
         """Return an AlbumInfo object for a bandcamp album page.
         If track url is given by mistake, find and fetch the album url instead.
         """
-        include_all = self.config["include_digital_only_tracks"]
         guru = self.guru(url, html=self._get(url))
-        return self.handle(partial(guru.album, include_all), url) if guru else None
+        return self.handle(guru, "album", url) if guru else None
 
     def get_track_info(self, url: str) -> Optional[TrackInfo]:
         """Returns a TrackInfo object for a bandcamp track page."""
         guru = self.guru(url)
-        return self.handle(partial(attrgetter("singleton"), guru), url) if guru else None
-
-    def handle(self, call: Callable, _id: str) -> Any:
-        try:
-            return call()
-        except (KeyError, ValueError, AttributeError):
-            self._info("Failed obtaining {}", _id)
-            return None
+        return self.handle(guru, "singleton", url) if guru else None
 
     def _search(self, query: str, search_type: str = ALBUM_SEARCH) -> Iterator[str]:
         """Return an iterator for item URLs of type search_type matching the query."""
@@ -285,10 +299,8 @@ class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
         while next_page_exists():
             self._info("Page {}", str(page))
             html = self._get(SEARCH_URL.format(query, page))
-            if not html:
-                break
 
-            for match in re.finditer(pattern, html):
+            for match in re.finditer(pattern, html or ""):
                 if len(urls) == max_urls:
                     break
                 url = match.groups()[0]
